/*
 * RTSP (Real Time Streaming Protocol) Reference Implementation
 * 
 *     Unix low-level network calls
 *
 *
 *  Copyright (C) 1996 Progressive Networks.
 * 
 *  This program is free software; you can only redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.  A copy of this
 *  license may be found in the "LICENSE" directory of this distribution.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY OR REPRESENTATION AS TO OWNERSHIP OR
 *  NON-INFRINGEMENT; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *  Please submit patches, modified versions and derivative works to
 *  Progressive Networks at "rtsp-feedback@prognet.com" or write to
 *  Progressive Networks, 1111 Third Ave. Ste 2900, Seattle, WA 98101, USA.
 *  We will do our best to incorporate your enhancements into the core
 *  distribution.
 *
 *  For more information about this program, please visit the following
 *  URL:
 *     http://www.real.com/prognet/rt 
 *

 *
 * $Id: socket.c,v 1.9 1997/04/01 17:31:36 robla Exp $
 *
 */

#include "socket.h"
#include "util.h"
#include "interface.h"

extern void terminate(int errorcode);

struct sockaddr peer;

SOCKET
tcp_open(struct sockaddr *name, int namelen)
{
    int f;

    if ((f = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
    	perror("socket");
      interface_error( "socket() error in tcp_open.\n" );
      terminate(-1);
    }

    if (connect(f, name, namelen) < 0)
    {
    	perror("connect");
      interface_error( "connect() error in tcp_open.\n" );
      terminate(-1);
    }

    return f;
}

SOCKET
tcp_listen(unsigned short port)
{
    SOCKET f;
    struct sockaddr_in s;

    if ((f = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
    	perror("socket");
      interface_error( "socket() error in tcp_listen.\n" );
      terminate(-1);
    }

    s.sin_family = AF_INET;
    s.sin_addr.s_addr = htonl(INADDR_ANY);
    s.sin_port = htons(port);

    if (bind (f, (struct sockaddr *)&s, sizeof (s)))
    {
    	perror ("bind");
      interface_error( "bind() error in tcp_listen.\n" );
      terminate(-1);
    }

    if (listen(f, 5) < 0)
    {
    	perror("listen");
      interface_error( "listen() error in tcp_listen.\n" );
      terminate(-1);
    }

    return f;
}

SOCKET
tcp_accept(SOCKET fd, struct sockaddr *addr, int *addrlen)
{
    SOCKET f;
    int namelen = sizeof (peer);

    if ((f = accept (fd, addr, addrlen)) < 0)
    {
    	perror ("accept");
      interface_error( "accept() error in tcp_accept.\n" );
      terminate(-1);
    }

    if (getpeername (f, &peer, &namelen))
    {
    	perror ("getpeername");
      interface_error( "getpeername() error in tcp_accept.\n" );
      terminate(-1);
    }

    return f;
}

int
tcp_read(SOCKET fd, void *buffer, int nbytes)
{
    int n;

    n = recv(fd, buffer, nbytes, 0);
#ifdef DUMP
    printf ("Read:\n");
    dump_buffer(buffer, n);
#endif

    return n;
}

int
tcp_write(SOCKET fd, void *buffer, int nbytes)
{
    int n;

    n = send(fd, buffer, nbytes, 0);
#ifdef DUMP
    printf ("Write:\n");
    dump_buffer(buffer, n);
#endif

    return n;
}

int
tcp_close(SOCKET fd)
{
    return closesocket(fd);
}

SOCKET
udp_open()
{
    SOCKET f;
    struct sockaddr_in s;
    int on = 1;

    if ((f = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
    	perror("socket");
      interface_error( "socket() error in udp_open.\n" );
    	terminate(-1);
    }

    s.sin_family = AF_INET;
    s.sin_addr.s_addr = htonl(INADDR_ANY);
    s.sin_port = htons(INADDR_ANY);

    if (bind (f, (struct sockaddr *)&s, sizeof (s)))
    {
    	perror ("bind");
      interface_error( "bind() error in udp_open.\n" );
      terminate(-1);
    }
    /* set to non-blocking */
    if (ioctlsocket(f, FIONBIO, &on) < 0)
    {
      perror("ioctl");
      interface_error( "ioctlsocket() error in udp_open.\n" );
      terminate(-1);
    }

    return f;
}

int
udp_close(SOCKET fd)
{
    return closesocket(fd);
}

int
udp_getport(SOCKET f, struct sockaddr_in *s)
{
   int namelen = sizeof(struct sockaddr_in);
   int err;

   if (getsockname (f, (struct sockaddr *) s, &namelen))
   {
      err = WSAGetLastError();
      perror ("getsockname");
      interface_error( "getsockname() error in udp_getport.\n" );
      terminate(-1);
   }
   return ntohs( s->sin_port );
}

SOCKET
udp_listen(unsigned short port)
{
    SOCKET f;
    struct sockaddr_in s;

    if ((f = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
    	perror("socket");
      interface_error( "socket() error in udp_listen.\n" );
      terminate(-1);
    }

    s.sin_family = AF_INET;
    s.sin_addr.s_addr = htonl(INADDR_ANY);
    s.sin_port = htons(port);

    if (bind (f, (struct sockaddr *)&s, sizeof (s)))
    {
    	perror ("bind");
      interface_error( "bind() error in udp_listen.\n" );
      terminate(-1);
    }

    return f;
}

SOCKET
mcast_open(struct sockaddr *name, int namelen)
{
    return INVALID_SOCKET;
}

SOCKET
mcast_listen(int port)
{
    return INVALID_SOCKET;
}

int
dgram_recvfrom(SOCKET s, void *buf, size_t len, int flags, struct sockaddr
	*from, int *fromlen)
{
    return recvfrom(s, buf, len, flags, from, fromlen);
}

int
dgram_sendto(SOCKET s, const void *msg, size_t len, int flags,
	const struct sockaddr *to, int tolen)
{
    return sendto(s, msg, len, flags, to, tolen);
}
